/***********************************************************************
*
*   FIBADD  
*   Lagged Fibonnacci random number generator using addition
*
*   Paul Coddington, May 1995.
*   Version 1.1
*
* General lagged Fibonnacci generator using subtraction (equivalent to
* addition), with lags p and q, i.e. F(p,q,-) in Marsaglia's notation.
* 
* The random numbers X_{i} are obtained from the sequence:
* 
*    X_{i} = X_{i-q} - X_{i-p}   mod M
*  
* where M is 1 if the X's are taken to be floating point reals in [0,1),
* as they are here.
* 
* For good results, the biggest lag should be at least 1000, and probably
* on the order of 10000.
* 
* The following lags give the maximal period of the generator, which is 
* (2^{p} - 1)2^{n-1} on integers mod 2^n or reals with n bits in the 
* mantissa (see Knuth, or W. Zerler, Information and Control, 15 (1969) 67, 
* for a complete more list).
* 
*     P     Q
*   9689   471 
*   4423  1393
*   2281   715
*   1279   418
*    607   273
*    521   168
*    127    63
* 
* This program is based on the implementation of RANMAR in 
* F. James, "A Review of Pseudo-random Number Generators", 
* Comput. Phys. Comm. 60, 329 (1990).
* 
* For more details, see:
* 
* D.E. Knuth, The Art of Computer Programming Vol. 2: 
* Seminumerical Methods, (Addison-Wesley, Reading, Mass., 1981).
* 
* P. L'Ecuyer, Random numbers for simulation, Comm. ACM 33:10, 85 (1990).
* 
* G.A. Marsaglia, A current view of random number generators,
* in Computational Science and Statistics: The Interface,
* ed. L. Balliard (Elsevier, Amsterdam, 1985).
* 
***********************************************************************/
 

#include <stdio.h>

#define MAXSEED 900000000

#define SIGBITS 24   /* Number of significant bits */

/* Lags */
#define P 1279
#define Q  418

/* Variables for lagged Fibonacci */
double u[P+1];  /* seed table */
int pt0, pt1;   /* pointers into the seed table */



/***************************************************************************
	Initialize the random number generator.
        Taken from RMARIN in James's review -- initializes
	every significant bit using a combination linear 
	congruential and small lag Fibonacci generator.
***************************************************************************/
void RandomSeed(long seed)
{
      int ij, kl, i, j, k, l;
      int ii, jj, m;
      double t, s;
      
      if (seed < 0) seed = - seed;
      seed = seed % MAXSEED;

      ij = seed / 30082;
      kl = seed - 30082 * ij;
      i = ((ij/177)% 177) + 2;
      j =  ij%177 + 2;
      k = ((kl/169)% 178) + 1;
      l =  kl%169;
      
      for ( ii = 1 ; ii <= P; ii++ ){
        s = 0.0;
        t = 0.5;
        for ( jj = 1 ; jj <= SIGBITS; jj++ ){
          m = (((i*j)% 179)*k)% 179;
          i = j;
          j = k;
          k = m;
          l = (53*l+1) % 169;
          if ( ((l*m)%64) >= 32) 
            s = s + t;
          t = 0.5 * t;
        }
        u[ii] = s;
      }
      pt0 = P;
      pt1 = Q;

      return;
}
  
/***************************************************************************
	Return a random double in [0.0, 1.0)
***************************************************************************/
double RandomNumber()
{
      double uni;
      
      uni = u[pt0] - u[pt1];
      if (uni < 0.0) uni = uni + 1.0;
      u[pt0] = uni;
      pt0 = pt0 - 1;
      if (pt0 == 0) pt0 = P;
      pt1 = pt1 - 1;
      if (pt1 == 0) pt1 = P;
            
      return(uni);
}


/***************************************************************************
*       Store the lag table and pointers.
***************************************************************************/
void WriteSeeds(seedname)
char seedname[128];
{
   FILE *f1,*fopen();
   int i,pointer[2];
   
   pointer[0] = pt0;
   pointer[1] = pt1;

   f1 = fopen(seedname,"w") ;
   if (f1 == NULL) {
     printf("Unable to open file %s\n", seedname); fflush(stdout);
     return -1;
   }
   fwrite(u[1],sizeof(double),P,f1);
   fflush(f1);
   fwrite(pointer,sizeof(int),2,f1);
   fclose(f1);
   return 0;
}


/***************************************************************************
*       Read in stored lag table and pointers.
***************************************************************************/
int ReadSeeds(seedname)
char seedname[128];
{
   FILE *f1,*fopen();
   int i,pointer[2];

   f1 = fopen(seedname,"r");
   if (f1 == NULL) {
     printf("Unable to open file %s\n", seedname); fflush(stdout);
     return -1;
   }
   fread(u[1],sizeof(double),P,f1);
   fread(pointer,sizeof(int),2,f1);
   fclose(f1);
   u[0] = 0.0;
   pt0 = pointer[0];
   pt1 = pointer[1];

   return 0;
}





